"""
    data source:
        https://www.hycom.org/dataserver/gofs-3pt1/analysis        

    web interface for manual hycom data retrieval:
        https://tds.hycom.org/thredds/dodsC/GLBy0.08/expt_93.0.html

    GLBy0.08/expt_93.0 
        Time range: Dec 4th, 2018 to  Present        
        Longitude convention: 0 to +360 degrees
        Depth coordinates are positive below the sea surface
        Times are in epoch hours since 2000-01-01 00:00 UTC

    API limitations:
        * Do not subset more than 1 day at a time using ncss.hycom.org
        * Do not use more than 2 concurrent connections per IP address when downloading data from ncss.hycom.org

"""

import logging
import requests
from functools import reduce
from datetime import datetime, timedelta

import numpy as np

from kadlu import index
from kadlu.geospatial.data_sources.data_util import (
    database_cfg,
    dt_2_epoch,
    fmt_coords,
    fmt_time,
    index_arr,
    logmsg,
    logmsg_nodata,
    storage_cfg,
    str_def,
)

# HYCOM data download URL
hycom_src = "http://tds.hycom.org/thredds/dodsC/GLBy0.08/expt_93.0"

# Specify which variable will be downloaded
hycom_tables = [
    'hycom_salinity', 'hycom_water_temp', 'hycom_water_u', 'hycom_water_v'
]

# Helper function
slices_str = lambda var, slices, steps=(
    1, 1, 1, 1
): f"{var}{''.join(map(lambda tup, step : f'[{tup[0]}:{step}:{tup[1]}]', slices, steps))}"


def initdb():
    """ Create tables in kadlu's geospatial.db database for storing HYCOM data"""
    conn, db = database_cfg()

    for var in hycom_tables:
        db.execute(f'CREATE TABLE IF NOT EXISTS {var}'
                   '( val     REAL NOT NULL,'
                   '  lat     REAL NOT NULL,'
                   '  lon     REAL NOT NULL,'
                   '  time    INT  NOT NULL,'
                   '  depth   INT  NOT NULL,'
                   '  source  TEXT NOT NULL )')

        db.execute(f'CREATE UNIQUE INDEX IF NOT EXISTS '
                   f'idx_{var} on {var}(time, lon, lat, depth, val, source)')
    conn.close()


def fetch_grid(**_):
    """ Download HYCOM lat/lon/time/depth arrays for grid indexing.
        
        Times are formatted as epoch hours since 2000-01-01 00:00.

        Returns:
            lat, lon, epoch, depth: numpy array
                The coordinate arrays
    """
    logger = logging.getLogger("kadlu")

    # fetch lat/lon arrays

    logger.info("[HYCOM] Fetching lat and lon indices ...")

    url = f"{hycom_src}.ascii?lat%5B0:1:4250%5D,lon%5B0:1:4499%5D"
    
    logger.debug(f"[HYCOM] data request URL: {url}")
    grid_netcdf = requests.get(url, timeout=450)
    assert (grid_netcdf.status_code == 200)

    meta, data = grid_netcdf.text.split\
    ("---------------------------------------------\n")
    lat_csv, lon_csv = data.split("\n\n")[:-1]
    lat = np.array(lat_csv.split("\n")[1].split(", "), dtype=float)
    lon = np.array(lon_csv.split("\n")[1].split(", "), dtype=float)

    logger.debug(f"[HYCOM] lon: {lon}")

    # fetch time array

    logger.info(f"[HYCOM] Fetching time index ...")

    url = f"{hycom_src}.ascii?time"
    logger.debug(f"[HYCOM] data request URL: {url}")
    time_netcdf = requests.get(url, timeout=450)
    assert (time_netcdf.status_code == 200)
    meta, data = time_netcdf.text.split\
    ("---------------------------------------------\n")
    csv = data.split("\n\n")[:-1][0]
    epoch = np.array(csv.split("\n")[1].split(', ')[1:], dtype=float)

    # fetch depth array

    logger.info(f"[HYCOM] Fetching depth index ...")

    depth = np.array([
        0.0, 2.0, 4.0, 6.0, 8.0, 10.0, 12.0, 15.0, 20.0, 25.0, 30.0, 35.0,
        40.0, 45.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0, 125.0, 150.0, 200.0,
        250.0, 300.0, 350.0, 400.0, 500.0, 600.0, 700.0, 800.0, 900.0, 1000.0,
        1250.0, 1500.0, 2000.0, 2500.0, 3000.0, 4000.0, 5000.0
    ])

    return lat, lon, epoch, depth


def load_grid():
    """ Put spatial grid into memory """
    with index(storagedir=storage_cfg(), bins=False, store=True) as fetchmap:
        return fetchmap(callback=fetch_grid, seed='hycom grids')[0]


class Hycom():
    """ Collection of module functions for fetching and loading HYCOM data.

        Attributes:
            lat, lon: arrays
                Lat/lon coordinates.
            epoch: array
                Time coordinates. Times are formatted as epoch hours since 2000-01-01 00:00
            depth: array
                Depth coordinates.
    """

    def __init__(self):
        self.ygrid, self.xgrid, self.epoch, self.depth = None, None, None, None
        self.logger = logging.getLogger("kadlu")
        self.name = "HYCOM"

    def load_salinity(self, **kwargs):
        return self.load_hycom('salinity', kwargs)

    def load_temp(self, **kwargs):
        return self.load_hycom('water_temp', kwargs)

    def load_water_u(self, **kwargs):
        return self.load_hycom('water_u', kwargs)

    def load_water_v(self, **kwargs):
        return self.load_hycom('water_v', kwargs)

    def callback(self, var, max_attempts=3, **kwargs):
        """ Builds indices for query, fetches data from HYCOM, and inserts into local database.

            Note: Null/NaN values are removed before the data is inserted into the local database.
            Null/NaN values occur when the grid overlaps with land or extends below the seafloor. 

            TODO: Add download progress bar, e.g., using the approach described here:
                https://stackoverflow.com/questions/37573483/progress-bar-while-download-file-over-http-with-requests

            Args:
                var: string
                    Variable to be fetched. complete list of variables here
                    https://tds.hycom.org/thredds/dodsC/GLBv0.08/expt_53.X/data/2015.html
                max_attempts: int
                    Maximum number of request attempts. Default is 3. Each request has a timeout of 120 s.
                kwargs: dict
                    boundaries as keyword arguments
        """
        # build request indexes
        south = kwargs["south"]
        north = kwargs["north"]
        west = kwargs["west"]
        east = kwargs["east"]
        top = kwargs["top"]
        bottom = kwargs["bottom"]
        start_epoch = dt_2_epoch(kwargs['start'])
        end_epoch = dt_2_epoch(kwargs['end'])

        haystack = np.array([self.epoch, self.depth, self.ygrid, self.xgrid], dtype=object)
        needles1 = np.array([start_epoch, top, south, west])
        needles2 = np.array([end_epoch, bottom, north, east])
        
        slices = list(
            zip(map(index_arr, needles1, haystack),
                map(index_arr, needles2, haystack))
        )
        
        assert reduce(
            np.multiply, map(lambda s: s[1] - s[0] + 1, slices)
        ) > 0, f"0 records available within query boundaries: {kwargs}"

        # generate data download request for HYCOM server
        url = f"{hycom_src}.ascii?{slices_str(var, slices)}"

        self.logger.debug(f"[{self.name}] data request URL: {url}")

        # make several download attempts until successful
        counter = 0
        timeout = 120
        code = None
        while counter < max_attempts:
            counter += 1
            self.logger.debug(f"[{self.name}] Requesting data from Hycom server ... (attempt no. {counter}/{max_attempts})")
            try:
                payload_netcdf = requests.get(url, stream=True, timeout=timeout, verify=True)
                code = payload_netcdf.status_code
                if code == 200:
                    debug_msg = f"[{self.name}] Data request was successful"
                    self.logger.debug(debug_msg)
                    break

                else:
                    debug_msg = f"[{self.name}] Could not access Hycom server; server returned status code {code}"
                    self.logger.debug(debug_msg)

            except requests.exceptions.ReadTimeout:
                debug_msg = f"[{self.name}] Request to Hycom server timed out"
                self.logger.debug(debug_msg)
 
        assert code == 200, f"[{self.name}] Data request unsuccesful. Could not access Hycom server."

        meta, data = payload_netcdf.text.split("---------------------------------------------\n")

        # parse response into numpy array
        arrs = data.split("\n\n")[:-1]
        shape_str, payload = arrs[0].split("\n", 1)
        shape = tuple(
            [int(x) for x in shape_str.split("[", 1)[1][:-1].split("][")])
        cube = np.ndarray(shape, dtype=float)

        for arr in payload.split("\n"):
            ix_str, row_csv = arr.split(", ", 1)
            a, b, c = [int(x) for x in ix_str[1:-1].split("][")]
            cube[a][b][c] = np.array(row_csv.split(", "), dtype=int)

        # build coordinate grid, populate values, adjust scaling, remove nulls
        flatten = reduce(np.multiply, map(lambda s: s[1] - s[0] + 1, slices))
        add_offset = 20 if 'salinity' in var or 'water_temp' in var else 0
        null_value = -10 if 'salinity' in var or 'water_temp' in var else -30

        grid = np.array([
            (None, y, x, t, d, 'hycom')
            for t in self.epoch[slices[0][0]:slices[0][1] + 1]
            for d in self.depth[slices[1][0]:slices[1][1] + 1]
            for y in self.ygrid[slices[2][0]:slices[2][1] + 1]
            for x in self.xgrid[slices[3][0]:slices[3][1] + 1]
        ])
        grid[:, 0] = np.reshape(cube, flatten) * 0.001 + add_offset
        grid = grid[grid[:, 0] != null_value]

        # insert into db
        initdb()
        conn, db = database_cfg()
        n1 = db.execute(f"SELECT COUNT(*) FROM hycom_{var}").fetchall()[0][0]
        db.executemany(
            f"INSERT OR IGNORE INTO hycom_{var} VALUES "
            "(?, ?, ?, CAST(? AS INT), CAST(? AS INT), ?)", grid)
        n2 = db.execute(f"SELECT COUNT(*) FROM hycom_{var}").fetchall()[0][0]
        db.execute("COMMIT")
        conn.commit()
        conn.close()

        logmsg('hycom', var, (n1, n2), **kwargs)
        return

    def fetch_hycom(self, var, kwargs, max_attempts=3):
        """ Fetch data from the HYCOM server.

            Kadlu's `index` class is used for 'binning' the data requests into 
            requests that span 1 degree in lat/lon, 24 hours in time (1 day), 
            and 0-5000 m in depth.

            A data request that spans multiple lat/lon degrees, multiple days, or 
            includes depths greater than 5000m is split into multiple such 'binned' 
            requests.

            Conversely, smaller data requests are 'inflated' to the size of one 'bin'.

            Args:
                var: string
                    Variable to be fetched. complete list of variables here
                    https://tds.hycom.org/thredds/dodsC/GLBv0.08/expt_53.X/data/2015.html
                kwargs: dict
                    boundaries as keyword arguments
                max_attempts: int
                    Maximum number of request attempts. Default is 3. Each request has a timeout of 120 s.
        """
        boundary_keys = ['west', 'east', 'south', 'north', 'top', 'bottom', 'start', 'end']

        kwargs = kwargs.copy()

        # input validation
        assert kwargs['start'] <= kwargs['end']
        assert kwargs['start'] > datetime(2018, 12, 4), f"[{self.name}] Historical data not available before 2018-12-04"
        assert kwargs['end'] < datetime.now(), f"[{self.name}] Forecast data not available"
        assert kwargs['south'] <= kwargs['north']
        assert kwargs['top'] <= kwargs['bottom']

        if self.ygrid is None:
            self.ygrid, self.xgrid, self.epoch, self.depth = load_grid()

        # convert longitudes from [-180;180] to [0;360]
        if kwargs['east'] < 0:
            kwargs['east'] = 360. + kwargs['east']

        if kwargs['west'] < 0:
            kwargs['west'] = 360. + kwargs['west']

        # if query spans meridian, make two seperate fetch requests
        if kwargs['west'] > kwargs['east']:
            self.logger.debug(f'[{self.name}] splitting request')
            argsets = [kwargs.copy(), kwargs.copy()]
            argsets[0]['east'] = self.xgrid[-1]
            argsets[1]['west'] = self.xgrid[0]
        else:
            argsets = [kwargs]

        for argset in argsets:
            args = {k: v for k, v in argset.items() if k in boundary_keys}

            self.logger.info(f'[{self.name}] Fetching {var} in region {fmt_coords(args)} for time period {fmt_time(args)}')

            with index(storagedir=storage_cfg(),
                       bins=True,
                       dx=1,
                       dy=1,
                       dz=5000,
                       dt=timedelta(hours=24),
                       **args) as fetchmap:
                fetchmap(callback=self.callback, var=var, max_attempts=max_attempts)

        return True

    def load_hycom(self, var, kwargs):
        """ Load HYCOM data from local database.

            If data is not present, attempts to fetch it from the HYCOM server.

            Although HYCOM uses the 0 to +360 degree longitude convention, the longitude 
            coordinates returned by this method adhere to the -180 to +180 convention 
            used everywhere else in Kadlu.

            Args:
                var:
                    Variable to be fetched. complete list of variables here
                    https://tds.hycom.org/thredds/dodsC/GLBv0.08/expt_53.X/data/2015.html
                south, north: float
                    ymin, ymax coordinate values. range: -90, 90
                kwargs: dict (boundaries as keyword arguments)
                    west, east: float
                        xmin, xmax coordinate values. range: -180, 180
                    start, end: datetime
                        temporal boundaries in datetime format

            Returns:
                values: array
                    values of the fetched variable
                lat: array
                    y grid coordinates
                lon: array
                    x grid coordinates
                epoch: array
                    timestamps in epoch hours since jan 1 2000
                depth: array
                    measured in meters
        """
        kwargs = kwargs.copy()

        # check if grids are initialized
        if self.ygrid is None:
            self.ygrid, self.xgrid, self.epoch, self.depth = load_grid()

        # convert from [-180;180] to [0;360]
        if kwargs['east'] < 0:
            kwargs['east'] = 360. + kwargs['east']

        if kwargs['west'] < 0:
            kwargs['west'] = 360. + kwargs['west']

        # recursive function call for queries spanning meridian
        if kwargs['west'] > kwargs['east']:
            kwargs1 = kwargs.copy()
            kwargs2 = kwargs.copy()
            kwargs1['west'] = self.xgrid[0]
            kwargs2['east'] = self.xgrid[-1]
            return np.hstack(
                (self.load_hycom(var, kwargs1), self.load_hycom(var, kwargs2)))

        # check for missing data
        self.fetch_hycom(var, kwargs, max_attempts=3)

        # validate and execute query
        assert 8 == sum(
            map(lambda kw: kw in kwargs.keys(), [
                'south', 'north', 'west', 'east', 'start', 'end', 'top',
                'bottom'
            ])), 'malformed query'

        assert kwargs['start'] <= kwargs['end']

        # prepare SQL query
        sql_query = ' AND '.join([
            f"SELECT * FROM hycom_{var} WHERE lat >= ?", 'lat <= ?',
            'lon >= ?', 'lon <= ?', 'time >= ?', 'time <= ?', 'depth >= ?',
            'depth <= ?', "source == 'hycom' "
        ]) + 'ORDER BY time, depth, lat, lon ASC'

        sql_values = [
                kwargs['south'], kwargs['north'], kwargs['west'], kwargs['east'],
                dt_2_epoch(kwargs['start']), dt_2_epoch(kwargs['end']), 
                kwargs['top'], kwargs['bottom']
        ]
        sql_values = tuple(map(str, sql_values))

        # query the local database
        conn, db = database_cfg()
        db.execute(sql_query, sql_values)
        rowdata = np.array(db.fetchall(), dtype=object).T

        conn.close()

        if len(rowdata) == 0:
            logmsg_nodata('hycom', var, **kwargs)
            return np.array([[], [], [], [], []])
        
        # convert longitude from [0;360] to [-180;180]
        lon = rowdata[2]
        lon[lon > 180] = lon[lon > 180] - 360
        rowdata[2] = lon

        return rowdata[0:5].astype(float)

    def load_water_uv(self, **kwargs):
        """ Load water speed, computed as sqrt(vu^2 + vv^2)

            If data is not present, attempts to fetch it from the HYCOM server.

            Args:
                var:
                    Variable to be fetched. complete list of variables here
                    https://tds.hycom.org/thredds/dodsC/GLBv0.08/expt_53.X/data/2015.html
                south, north: float
                    ymin, ymax coordinate values. range: -90, 90
                kwargs: dict (boundaries as keyword arguments)
                    west, east: float
                        xmin, xmax coordinate values. range: -180, 180
                    start, end: datetime
                        temporal boundaries in datetime format

            Returns:
                values: array
                    Water speed values, in m/s
                lat: array
                    y grid coordinates
                lon: array
                    x grid coordinates
                epoch: array
                    timestamps in epoch hours since jan 1 2000
                depth: array
                    measured in meters
        """
        kwargs = kwargs.copy()

        # convert from [-180;180] to [0;360]
        if kwargs['east'] < 0:
            kwargs['east'] = 360. + kwargs['east']

        if kwargs['west'] < 0:
            kwargs['west'] = 360. + kwargs['west']

        self.fetch_hycom('water_u', kwargs, max_attempts=3)
        self.fetch_hycom('water_v', kwargs, max_attempts=3)

        sql_query = ' AND '.join(['SELECT hycom_water_u.val, hycom_water_u.lat, hycom_water_u.lon, hycom_water_u.time, hycom_water_u.depth, hycom_water_v.val FROM hycom_water_u '\
                'INNER JOIN hycom_water_v '\
                'ON hycom_water_u.lat == hycom_water_v.lat',
                'hycom_water_u.lon == hycom_water_v.lon',
                'hycom_water_u.time == hycom_water_v.time '\
                'WHERE hycom_water_u.lat >= ?',
                'hycom_water_u.lat <= ?',
                'hycom_water_u.lon >= ?',
                'hycom_water_u.lon <= ?',
                'hycom_water_u.time >= ?',
                'hycom_water_u.time <= ?',
                'hycom_water_u.depth >= ?',
                'hycom_water_u.depth <= ?',
            ]) + ' ORDER BY hycom_water_u.time, hycom_water_u.lat, hycom_water_u.lon ASC'

        sql_values = [
                kwargs['south'], kwargs['north'], kwargs['west'], kwargs['east'],
                dt_2_epoch(kwargs['start']), dt_2_epoch(kwargs['end']), 
                kwargs['top'], kwargs['bottom']
        ]
        sql_values = tuple(map(str, sql_values))

        # query the local database
        conn, db = database_cfg()
        db.execute(sql_query, sql_values)
        qry = np.array(db.fetchall()).T

        conn.close()

        if len(qry) == 0:
            self.logger.warning(f'[{self.name}] water_uv: no data found in region '
                            f'{fmt_coords(kwargs)}, returning empty arrays')
            return np.array([[], [], [], [], []])

        water_u, lat, lon, epoch, depth, water_v = qry
        val = np.sqrt(np.square(water_u) + np.square(water_v))

        # convert longitude from [0;360] to [-180;180]
        lon[lon > 180] = lon[lon > 180] - 360

        return np.array((val, lat, lon, epoch, depth)).astype(float)

    def __str__(self):
        info = '\n'.join([
            "Native hycom .[ab] data converted to NetCDF at the Naval",
            "Research Laboratory, interpolated to 0.08° grid between",
            "40°S-40°N (0.04° poleward) containing 40 z-levels.",
            "Availability: 2014-July to Present",
            "\thttps://www.hycom.org/data/glby0pt08"
        ])

        args = ("(south, north, west, east, "
                "start, end, top, bottom)")

        return str_def(self, info, args)



import pytest
import kadlu
import numpy as np
from datetime import datetime
from kadlu.geospatial.data_sources.hycom import Hycom
from kadlu.geospatial.data_sources.data_util import reshape_4D


@pytest.mark.hycom_access
def test_fetch_load_salinity():
    """ Check that we can load salinity for a small region across the anti-meridian 
        down to 100 meters depth.       
    
        Note: If the data already exists in the local database instance, it will not 
        be re-downloaded.
    """
    south, west = 44, 179
    north, east = 45, -179
    top, bottom = 0, 100
    start = datetime(2020, 1, 10)
    end = datetime(2020, 1, 10, 12)

    data = kadlu.load(
        var='salinity',
        source='hycom',
        south=south,
        north=north,
        west=west,
        east=east,
        start=start,
        end=end,
        top=top,
        bottom=bottom
    )

    data = reshape_4D(data)

    # check that lat/lon coordinates are what they should be
    lats_answ = np.arange(44.0, 45.01, 0.04)
    assert np.all(np.isclose(data["lats"], lats_answ))
    lons_answ = np.concatenate([np.arange(-179.92, -179.03, 0.08), np.arange(179.04, 180.01, 0.08)])
    assert np.all(np.isclose(data["lons"], lons_answ))

    # check that salinity is 33 +- 1
    assert np.all(np.isclose(data["values"], 33, atol=1))


@pytest.mark.hycom_access
def test_load_water_uv():
    """ Check that we can load water speeds for a small region of the Atlantic ocean
        down to 500 m depth.
    
        Note: If the data already exists in the local database instance, it will not 
        be re-downloaded.
    """
    start = datetime(2020, 1, 10)
    end = datetime(2020, 1, 10, 12)

    bounds = dict(
        south=44.01,
        north=44.31,
        west=-63.1,
        east=-62.9,
        top=0,
        bottom=500,
        start=start,
        end=end,
    )

    data = kadlu.load(source='hycom', var='water_uv', **bounds)

    assert data.shape == (5, 34780)

    data = reshape_4D(data)

    # check that lat/lon,depth,epoch coordinates are what they should be

    lats_answ = np.array([44.04000092, 44.08000183, 44.11999893, 44.15999985, 44.20000076, 44.24000168, 44.27999878])
    assert np.all(np.isclose(data["lats"], lats_answ, rtol=1E-6))

    lons_answ = np.array([-63.03997803, -62.96002197])
    assert np.all(np.isclose(data["lons"], lons_answ, rtol=1E-6))

    depths_answ = np.array([  0.,   2.,   4.,   6.,   8.,  10.,  12.,  15.,  20.,  25.,  30., 35.,  40.,  45.,  50.,  60.,  70.,  80.,  90., 100., 125., 150., 200.])
    assert np.all(np.isclose(data["depths"], depths_answ, rtol=1E-6))

    times_answ = np.array([175536., 175539., 175542., 175545., 175548.])
    assert np.all(np.isclose(data["times"], times_answ, rtol=1E-6))


@pytest.mark.hycom_access
def test_load_water_u():
    """ Check that we can load water speed u component for a small region of the Atlantic ocean
        down to 500 m depth.

        Note: If the data already exists in the local database instance, it will not 
        be re-downloaded.
    """
    start = datetime(2020, 1, 10)
    end = datetime(2020, 1, 10, 12)

    bounds = dict(
        south=44.01,
        north=44.31,
        west=-63.1,
        east=-62.9,
        top=0,
        bottom=500,
        start=start,
        end=end,
    )

    H = Hycom()

    H.fetch_hycom(var="water_u", max_attempts=4, kwargs=bounds)

    data = H.load_water_u(**bounds)

    assert len(data) > 0  #check that something was loaded


'''

    TODO: finish implementing below test to check that 
            there is no water_u data on land

def test_no_data_on_land():

    start = datetime(2020, 1, 10)
    end = datetime(2020, 1, 10, 12)

    # west coast of Olympic peninsula, WA
    bounds = dict(
        south=47.5,
        north=48.4,
        west=-125,
        east=-124,
        top=0,
        bottom=500,
        start=start,
        end=end,
    )

    H = Hycom()
    data = H.load_water_u(**bounds)

    lat = data[1]
    lon = data[2]
    import matplotlib.pyplot as plt
    plt.scatter(lon, lat)
    plt.show()
'''



"""
    data fetching utils, function maps, and constant variables
    collection of objects
"""

from datetime import datetime

#import kadlu.geospatial.data_sources.chs as chs
import kadlu.geospatial.data_sources.era5 as era5
import kadlu.geospatial.data_sources.gebco as gebco
import kadlu.geospatial.data_sources.hycom as hycom
import kadlu.geospatial.data_sources.wwiii as wwiii
#import kadlu.geospatial.data_sources.nemo as nemo

# dict for mapping strings to callback functions
load_map = dict(
    #        bathy_chs           = chs  .Chs()  .load_bathymetry,
    #temp_hycom          = hycom.Hycom().load_temp,
    temperature_hycom=hycom.Hycom().load_temp,
    salinity_hycom=hycom.Hycom().load_salinity,
    water_uv_hycom=hycom.Hycom().load_water_uv,
    water_u_hycom=hycom.Hycom().load_water_u,
    water_v_hycom=hycom.Hycom().load_water_v,
    wavedir_era5=era5.Era5().load_wavedirection,
    wavedirection_era5=era5.Era5().load_wavedirection,
    waveheight_era5=era5.Era5().load_windwaveswellheight,
    waveperiod_era5=era5.Era5().load_waveperiod,
    insolation_era5=era5.Era5().load_insolation,
    irradiance_era5=era5.Era5().load_irradiance,
    wind_uv_era5=era5.Era5().load_wind_uv,
    wind_u_era5=era5.Era5().load_wind_u,
    wind_v_era5=era5.Era5().load_wind_v,
    precipitation_era5=era5.Era5().load_precipitation,
    snowfall_era5=era5.Era5().load_snowfall,
    flux_ocean_era5=era5.Era5().load_flux_ocean,
    flux_waves_era5=era5.Era5().load_flux_waves,
    stress_ocean_era5=era5.Era5().load_stress_ocean,
    #precip_type_era5    = era5 .Era5() .load_precip_type,
    wavedir_wwiii=wwiii.Wwiii().load_wavedirection,
    wavedirection_wwiii=wwiii.Wwiii().load_wavedirection,
    waveheight_wwiii=wwiii.Wwiii().load_windwaveheight,
    waveperiod_wwiii=wwiii.Wwiii().load_waveperiod,
    wind_uv_wwiii=wwiii.Wwiii().load_wind_uv,
    wind_u_wwiii=wwiii.Wwiii().load_wind_u,
    wind_v_wwiii=wwiii.Wwiii().load_wind_v,
    #bathy_gebco         = gebco.Gebco().load_bathymetry,
    bathymetry_gebco=gebco.Gebco().load_bathymetry,
    #water_u_nemo        = nemo .Nemo().load_uvelocity,
    #water_v_nemo        = nemo .Nemo().load_vvelocity,
    #water_w_nemo        = nemo .Nemo().load_wvelocity,
    #salinity_nemo       = nemo .Nemo().load_salinity,
    #temp_nemo           = nemo .Nemo().load_temperature,
    #turbidity_nemo      = nemo .Nemo().load_fraserturbidity,
)

precip_type_era5_map = {
    0: 'Reserved',
    1: 'Rain',
    2: 'Thunderstorm',
    3: 'Freezing rain',
    4: 'Mixed/Ice',
    5: 'Snow',
    6: 'Wet snow',
    7: 'Mixture of rain and snow',
    8: 'Ice pellets',
    9: 'Graupel',
    10: 'Hail',
    11: 'Drizzle',
    12: 'Freezing drizzle',
    255: 'Missing',
}
_ = [precip_type_era5_map.update({i: 'Reserved'}) for i in range(13, 192)]
_ = [
    precip_type_era5_map.update({i: 'Reserved for local use'})
    for i in range(192, 255)
]

precip_type_map = dict(era5=(era5.Era5().load_precip_type,
                             precip_type_era5_map))

default_val = dict(south=44.25,
                   west=-64.5,
                   north=44.70,
                   east=-63.33,
                   top=0,
                   bottom=5000,
                   start=datetime(2015, 3, 1),
                   end=datetime(2015, 3, 1, 12))

var3d = (
    'temp',
    'temperature',
    'salinity',
    'water_u',
    'water_v',
    'water_uv',
)

source_map = ("""
    CHS   (Canadian Hydrography Service)
          bathymetry:       bathymetric data in Canada's waterways. metres, variable resolution \n
    ERA5  (Global environmental dataset from Copernicus Climate Data Store)
          wavedir:          mean wave direction, degrees
          waveheight:       combined height of wind, waves, and swell. metres
          waveperiod:       mean wave period, seconds
          wind_uv:          wind speed computed as sqrt(u^2 + v^2) / 2, where u, v are direction vectors
          wind_u:           wind speed coordinate U-vector, m/s
          wind_v:           wind speed coordinate V-vector, m/s
          precipitation:
          precip_type:
          snowfall:
          flux_ocean:
          flux_waves:
          stress_ocean: 
          insolation:       hourly accumulated solar downwards radiation, J/m^2
          irradiance:       hourly mean solar downwards irradiance, W/m^2\n
    GEBCO (General Bathymetric Chart of the Oceans)
          bathymetry:       global bathymetric and topographic data. metres below sea level \n
    HYCOM (Hybrid Coordinate Ocean Model)
          salinity:         g/kg salt in water
          temp:             degrees celsius
          water_uv:         ocean current computed as sqrt(u^2 + v^2) / 2, where u, v are direction vectors
          water_u:          ocean current coordinate U-vector, m/s
          water_v:          ocean current coordinate V-vector, m/s \n
    WWIII (WaveWatch Ocean Model Gen 3)
          wavedir:          primary wave direction, degrees
          waveheight:       combined height of wind and waves, metres
          waveperiod:       primary mean wave period, seconds
          wind_uv:          wind speed computed as sqrt(u^2 + v^2) / 2, where u, v are direction vectors
          wind_u:           wind speed coordinate U-vector, m/s
          wind_v:           wind speed coordinate V-vector, m/s \n
    NEMO  (3D zonal component fields from the SalishSeaCoast Model)
          water_u:
          water_v:
          water_w:
          salinity:
          temp:
          turbidity:
          mesozooplankton:
          microzooplankton:

    """)


class SourceMap():

    #def fetch_map(self):
    #    return fetch_map

    def load_map(self):
        return load_map

    def default_val(self):
        return default_val

    def var3d(self):
        return var3d

    def source_map(self):
        return source_map



''' job scheduler and hashmap database utility '''

import os
import json
import pickle
import sqlite3
import logging
from hashlib import md5
from multiprocessing import Pool
from datetime import datetime, timedelta

from numpy import arange


class index():
    ''' database and process management utility

        An SQL database for caching arbitrary binary.  Intended to be used as a
        wrapper for arbitrary function calls, where the result of the function
        is stored as a binary BLOB in the db. A hash of the function name,
        path, and input arguments will be interpreted as a 16-bit integer to be
        used as primary key on a clustered index.

        Optionally can also store True/False instead of binary, segment
        a bounding-box coordinate region into smaller partitions, and
        run tasks in parallel
    '''
    def hash_seed(self, callback, passkwargs={}):
        ''' Each data blob in the hashmap table is assumed to be uniquely
            described by a function and its input arguments.
            The returned string is a serializeable representation of the
            module path, function name, and arguments supplied to the function.
            This allows, for example, the output of any function to be hashed
            before actually calling the function.
            A common use case is caching the results of long-running functions.
        '''
        return (f'{callback.__module__}.{callback.__name__}'
                f':{json.dumps(passkwargs, default=str, sort_keys=True)}')

    def hash_dict(self, kwargs, seed=''):
        ''' compute 64-bit integer hash for a given dictionary.
            sqlite maxint: +/- (2^63)-1
        '''
        text = str(seed) + json.dumps(kwargs, sort_keys=True, default=str)
        checksum = md5((text).encode('utf-8')).hexdigest()
        return (int(checksum, base=16) >> 64) - (2**63) - 1

    def _spacebins(self, a, b, delta):
        ''' returns an array of evenly spaced values between a and b with a
            step size of delta.
            a modulus is applied to shift start and end values to the nearest
            integer outside the bounds of (a,b).

            this function ensures deterministic results when bins=True,
            with time/coordinate bounds as (a,b) values, and delta being one
            of (dx, dy, dz, dt).
        '''
        return arange(
            min(a, b) - (min(a, b) % (delta * 1)),
            max(a, b) - (max(a, b) % (delta * -1)), delta)

    def __init__(self,
                 *,
                 pool=1,
                 store=False,
                 inmemory=False,
                 bins=True,
                 dx=2,
                 dy=2,
                 dz=5000,
                 dt=timedelta(days=1),
                 storagedir=os.getcwd(),
                 filename='checksums.db',
                 **kwargs):
        """
            args:
                pool:
                    number of processes to run in parallel. when not using the
                    parallelindex subclass, this must be equal to 1.
                store:
                    determines whether to store the results of the callback.
                    when False, nothing will be serialized. only a hash of the
                    function and input arguments will be stored to determine
                    whether the callback was called
                inmemory:
                    when store=True and inmemory=True, callback results can be
                    stored to an in-memory database. this allows for faster
                    read and update times for processing that requires frequent
                    updates. will be copied to disk upon __exit__
                bins:
                    if True, boundary kwargs supplied to __init__ will be
                    split into multiple smaller area bins and passed to the
                    callback function. when
                dx:
                    delta longitude bin size (float)
                dy:
                    delta latitude bin size (float)
                dz:
                    delta depth bin size (float)
                dt:
                    delta time bin size (timedelta)
                storagedir:
                    directory to store spatial hashes and binary objects
                kwargs:
                    boundary arguments to split into bins, e.g.
                    (south, west, north, east, top, bottom, start, end)
        """
        assert kwargs != {} or not bins, 'no boundaries provided'
        self.name = self.__class__.__name__
        self.logger = logging.getLogger("kadlu")
        self.store = store
        self.pool = pool
        self.storagedir = storagedir
        self.inmemory = inmemory
        self.storage = (os.path.join(storagedir, filename)
                        if not inmemory else ':memory:')
        self.kwargslist = list(self.bin_kwargs(
            dx,
            dy,
            dz,
            dt,
            **kwargs,
        )) if bins else [kwargs]

    def __enter__(self):
        assert self.kwargslist != [], 'empty kwargs!'
        assert os.path.isdir(str(
            self.storagedir)), f'invalid dir {self.storagedir}'

        with sqlite3.connect(self.storage) as con:
            db = con.cursor()
            db.execute('''
                CREATE TABLE IF NOT EXISTS
                hashmap(
                    hash INTEGER PRIMARY KEY,
                    bytes BLOB
                )
                WITHOUT ROWID;''')
            db.execute('CREATE UNIQUE INDEX IF NOT EXISTS '
                       'idx_map on hashmap(hash)')
            zeros = ''.join(['0' for _ in range(32)])
            ones = ''.join(['f' for _ in range(32)])
            minval = (int(zeros, base=16) >> 64) - (2**63)
            maxval = (int(ones, base=16) >> 64) - (2**63)
            db.execute('INSERT OR IGNORE INTO hashmap VALUES (?,?)',
                       (minval, pickle.dumps(None)))
            db.execute('INSERT OR IGNORE INTO hashmap VALUES (?,?)',
                       (maxval, pickle.dumps(None)))
        # TODO:
        # if self.inmemory: copy disk database into memory
        # https://stackoverflow.com/questions/4019081/how-to-copy-a-sqlite-table-from-a-disk-database-to-a-memory-database-in-python
        assert not self.inmemory, 'feature not yet implemented'
        return self

    def __call__(self, *, callback, **passkwargs):
        return list(self.__call_generator__(callback=callback, **passkwargs))

    def __call_generator__(self, *, callback, **passkwargs):
        seed = self.hash_seed(callback, passkwargs)
        assert self.pool == 1, 'use parallelindex for processing pool'

        for kwargs in self.kwargslist:
            if not self.serialized(kwargs, seed):
                self.insert_hash(kwargs, seed,
                                 callback(**passkwargs, **kwargs))
            elif self.inmemory:
                self.update_hash(kwargs, seed,
                                 callback(**passkwargs, **kwargs))
            yield pickle.loads(self.serialized(kwargs, seed))

    def __exit__(self, exc_type, exc_value, tb):
        # TODO:
        # if self.inmemory: copy in-memory database back to disk
        # https://stackoverflow.com/questions/4019081/how-to-copy-a-sqlite-table-from-a-disk-database-to-a-memory-database-in-python
        assert not self.inmemory, 'feature not yet implemented'

    def insert_hash(self, kwargs={}, seed='', obj=None):
        """ create a checksum of the arguments in the hashmap

            args:
                kwargs:
                    dict containing parameters to be hashed,
                    e.g. serializeable function arguments
                seed:
                    additional salt added to the hash
                obj:
                    arbitrary binary data that will be attributed to the hash
                    in the hashmap db table
        """
        self.logger.debug(f'[{self.name}] insert hash {self.hash_dict(kwargs, seed)}\n'
                      f'[{self.name}] seed: {seed}\n'
                      f'[{self.name}] bin: kwargs = {kwargs}')
        with sqlite3.connect(self.storage) as con:
            db = con.cursor()
            db.execute(
                'INSERT INTO hashmap VALUES (?,?)',
                (self.hash_dict(kwargs, seed),
                 bytes(
                     pickle.dumps(obj) if self.store else pickle.dumps(None))))

    def update_hash(self, kwargs={}, seed='', obj=None):
        """ update hashmap checksum value

            args:
                kwargs:
                    dict containing parameters to be hashed,
                    e.g. serializeable function arguments
                seed:
                    additional salt added to the hash
                obj:
                    arbitrary binary data that will be attributed to the hash
                    in the hashmap db table
        """
        self.logger.debug(f'[{self.name}] update hash {self.hash_dict(kwargs, seed)}\n'
                      f'[{self.name}] seed: {seed}\n'
                      f'[{self.name}] bin: kwargs = {kwargs}')
        with sqlite3.connect(self.storage) as con:
            db = con.cursor()
            db.execute('UPDATE hashmap SET bytes = ? WHERE hash = ?',
                       (pickle.dumps(obj), self.hash_dict(kwargs, seed)))

    def drop_hash(self, kwargs={}, seed=''):
        """
            args:
                kwargs:
                    dict containing parameters to be hashed,
                    e.g. serializeable function arguments
                seed:
                    additional salt added to the hash
        """
        self.logger.debug(f'[{self.name}] drop hash {self.hash_dict(kwargs, seed)}\n'
                      f'[{self.name}] seed: {seed }\n'
                      f'[{self.name}] bin: kwargs = {kwargs}')
        with sqlite3.connect(self.storage) as con:
            db = con.cursor()
            db.execute('DELETE FROM hashmap WHERE hash = ?',
                       (self.hash_dict(kwargs, seed), ))

    def serialized(self, kwargs={}, seed=''):
        """
            args:
                kwargs:
                    dict containing parameters to be hashed,
                    e.g. serializeable function arguments
                seed:
                    additional salt added to the hash

            returns:
                binary object or True if hash exists in database, else False
        """
        with sqlite3.connect(self.storage) as con:
            db = con.cursor()
            db.execute('SELECT * FROM hashmap WHERE hash == ?',
                       (self.hash_dict(kwargs, seed), ))
            res = db.fetchone()

        self.logger.debug(f'[{self.name}] check hash {self.hash_dict(kwargs, seed)}'
                      f': {"exists" if res is not None else "missing" }\n'
                      f'[{self.name}] seed: {seed}\n'
                      f'[{self.name}] bin: kwargs = {kwargs}')

        if res is None:
            return False
        if res[1] is None:
            return True
        if res[1] is not None:
            return res[1]

    def bin_kwargs(self, dx, dy, dz, dt, **kwargs):
        """ generate argument sets as area subsets of boundary kwargs

            kwargs are split into dx° * dy° * dz° * dt bins.
            coordinates are rounded to nearest outer-boundary degree integer

            args:
                deltas: (float or timedelta)
                    bin size of axes

                **kwargs (dict)
                    boundary arguments, e.g.

            >>> kwargs = {
            ...     'south': 43.21, 'north': 46.54,
            ...     'west': -123.45, 'east': -110.01,
            ...     'top': 0, 'bottom': 5000,
            ...     'start': datetime(2000, 1, 1), 'end': datetime(2000, 1, 2)
            ... }

            yields:
                dictionary containing args as a subset of input boundaries
        """
        if 'start' not in kwargs.keys():
            kwargs['start'] = datetime(2000, 1, 1)
        if 'end' not in kwargs.keys():
            kwargs['end'] = kwargs['start'] + dt

        for axmin, axmax, delta in zip(('west', 'south', 'bottom', 'start'),
                                       ('east', 'north', 'top', 'end'),
                                       (dx, dy, dz, dt)):
            if axmin not in kwargs.keys():
                kwargs[axmin] = 0
            if axmax not in kwargs.keys():
                kwargs[axmax] = 0
            if min(kwargs[axmin],
                   kwargs[axmax]) == max(kwargs[axmin], kwargs[axmax]):
                kwargs[axmax] += delta

        for x in self._spacebins(kwargs['west'], kwargs['east'], dx):

            for y in self._spacebins(kwargs['south'], kwargs['north'], dy):

                for z in self._spacebins(kwargs['top'], kwargs['bottom'], dz):

                    for t in arange(kwargs['start'].date(), kwargs['end'],
                                    dt).astype(datetime):

                        yield dict(
                            zip((
                                'west',
                                'east',
                                'south',
                                'north',
                                'top',
                                'bottom',
                                'start',
                                'end',
                            ), (
                                x,
                                x + dx,
                                y,
                                y + dy,
                                z,
                                z + dz,
                                t,
                                t + dt,
                            )))


class parallelindex(index):
    """ run index jobs in a parallel processing pool """
    def __call__(self, *, callback, **passkwargs):
        assert len(
            self.kwargslist) > 1, 'nothing to parallelize when bins=False'
        with Pool(self.pool) as p:
            return list(
                p.map(
                    self.__call_generator__,
                    zip(
                        (callback for _ in self.kwargslist),
                        self.kwargslist,
                        (passkwargs for _ in self.kwargslist),
                    ),
                ))

    def __call_generator__(self, args):
        callback, kwargs, passkwargs = args
        seed = self.hash_seed(callback, passkwargs)

        if not self.serialized(kwargs, seed):
            self.insert_hash(kwargs, seed, callback(**passkwargs, **kwargs))

        elif self.inmemory:
            self.update_hash(kwargs, seed, callback(**passkwargs, **kwargs))

        return pickle.loads(self.serialized(kwargs, seed))


'''
class memindex(index):
    pass
'''



import os
import logging
import zipfile
import requests
from datetime import datetime, timedelta

import numpy as np
from tqdm import tqdm

import kadlu
from kadlu import index
from kadlu.geospatial.data_sources.data_util import (
    database_cfg,
    logmsg,
    logmsg_nodata,
    storage_cfg,
)

fname = 'GEBCO_2021.nc'
url = 'https://www.bodc.ac.uk/data/open_download/gebco/gebco_2021/zip/'


def initdb():
    conn, db = database_cfg()
    db.execute('CREATE TABLE IF NOT EXISTS gebco'
               '(   val     REAL    NOT NULL,  '
               '    lat     REAL    NOT NULL,  '
               '    lon     REAL    NOT NULL  )')
    db.execute(
        'CREATE UNIQUE INDEX IF NOT EXISTS idx_gebco on gebco(val, lat, lon)')
    conn.close()


class Gebco():

    def fetch_bathymetry_grid(self):
        """ download netcdf archive and extract it """

        zipf = os.path.join(kadlu.storage_cfg(), "gebco_2021_netcdf.zip")

        # unzip the file if necessary
        if not os.path.isfile(os.path.join(storage_cfg(), fname)):

            # if there is no zip file, download it
            if not os.path.isfile(zipf):
                logging.info('downloading gebco bathymetry...')
                with requests.get(url, stream=True) as payload_netcdf:
                    assert payload_netcdf.status_code == 200, 'error fetching file'
                    with open(zipf, 'wb') as f:
                        with tqdm(total=4011413504,
                                  desc='gebco_2021_netcdf.zip',
                                  unit='B',
                                  unit_scale=True) as t:
                            for chunk in payload_netcdf.iter_content(
                                    chunk_size=8192):
                                f.write(chunk)
                                _ = t.update(8192)

            # unzip the downloaded file
            with zipfile.ZipFile(zipf, 'r') as zip_ref:
                logging.info('extracting bathymetry data...')
                zip_ref.extractall(path=storage_cfg())
        return

    def fetch_callback(self,
                       south,
                       north,
                       west,
                       east,
                       top=None,
                       bottom=None,
                       start=None,
                       end=None):
        """ build data grid indexes from .nc file and insert into database """

        if not os.path.isfile(os.path.join(storage_cfg(), fname)):
            self.fetch_bathymetry_grid()

        rows = kadlu.load_file(os.path.join(storage_cfg(), fname),
                               south=south,
                               north=north,
                               west=west,
                               east=east).T

        initdb()
        conn, db = database_cfg()
        n1 = db.execute("SELECT COUNT(*) FROM gebco ").fetchall()[0][0]
        db.executemany(
            "INSERT OR IGNORE INTO gebco (val, lat, lon)  VALUES (?,?,?)",
            rows)
        n2 = db.execute("SELECT COUNT(*) FROM gebco ").fetchall()[0][0]

        logmsg('gebco',
               'bathymetry', (n1, n2),
               south=south,
               west=west,
               north=north,
               east=east)
        conn.commit()
        conn.close()
        return

    def load_bathymetry(self, south, north, west, east, **_):
        """ load gebco bathymetry data """

        with index(dx=2,
                   dy=2,
                   dz=99999,
                   dt=timedelta(hours=24),
                   storagedir=storage_cfg(),
                   bins=True,
                   south=south,
                   north=north,
                   west=west,
                   east=east,
                   top=0,
                   bottom=99999,
                   start=datetime(2000, 1, 1),
                   end=datetime(2000, 1, 2)) as fetchmap:
            _ = fetchmap(callback=self.fetch_callback)

        conn, db = kadlu.database_cfg()
        qry = ' AND '.join([
            'SELECT val, lat, lon FROM gebco WHERE lat >= ?', 'lat <= ?',
            'lon >= ?', 'lon <= ?'
        ])
        logging.debug(f'query: {qry}')
        db.execute(qry, (south, north, west, east))
        res = np.array(db.fetchall()).T
        conn.close()
        if len(res) == 0:
            logmsg_nodata('gebco',
                          'bathymetry',
                          south=south,
                          north=north,
                          west=west,
                          east=east)
            return np.array([[], [], []])
        val, lat, lon = res
        return val, lat, lon



import os
import logging
import pytest
import numpy as np
from datetime import datetime

from kadlu.geospatial.ocean import Ocean

gitlab = os.environ.get("GITLAB_CI") is not None

bounds = dict(
    start=datetime(2019, 1, 9),
    end=datetime(2019, 1, 9, 3),
    south=42,
    west=-62.5,
    north=44,
    east=-60.5,
    top=0,
    bottom=5000
)

test_lat, test_lon, test_depth = bounds['south'], bounds['west'], bounds['top']


def test_null_ocean():
    """ Test that ocean is initialized with all variables set to null (0)"""
    o = Ocean(load_bathymetry=0, **bounds)
    assert o.bathymetry(lat=test_lat, lon=test_lon) == 0
    assert o.temperature(lat=test_lat, lon=test_lon, depth=test_depth) == 0
    assert o.salinity(lat=test_lat, lon=test_lon, depth=test_depth) == 0
    assert o.wavedir(lat=test_lat, lon=test_lon) == 0
    assert o.waveheight(lat=test_lat, lon=test_lon) == 0
    assert o.waveperiod(lat=test_lat, lon=test_lon) == 0
    assert o.wind_uv(lat=test_lat, lon=test_lon) == 0
    assert o.origin == (43.0, -61.5)
    assert o.boundaries == bounds


def test_uniform_bathy():
    """ Test that ocean can be initialized with uniform bathymetry"""
    o = Ocean(load_bathymetry=500.5, **bounds)
    assert o.bathymetry(lat=test_lat, lon=test_lon) == 500.5
    assert o.temperature(lat=test_lat, lon=test_lon, depth=test_depth) == 0


def test_interp_uniform_temp():
    """ Test that we can interpolate a uniform ocean temperature on any set of coordinates"""
    o = Ocean(load_temperature=16.1, **bounds)
    assert o.temperature(lat=41.2, lon=-66.0, depth=-33.0) == 16.1


def test_uniform_bathy_deriv():
    """ Test that uniform bathy has derivative zero"""
    o = Ocean(load_bathymetry=-500.5, **bounds)
    assert o.bathymetry(lat=1, lon=17, dlon=1) == 0


def test_load_as_array_on_irregular_grid():
    """ Test that ocean can be initialized with data sampled on an irregular grid, passed as arrays
        Also test that we can retrieve interpolated data.
    """
    # create 100 data points with bathymetry and salinity data sampled irregularly within the bounding box
    n = 100
    lats = np.random.uniform(low=bounds["south"], high=bounds["north"], size=n)
    lons = np.random.uniform(low=bounds["west"], high=bounds["east"], size=n)
    depths = np.random.uniform(low=bounds["top"], high=bounds["bottom"], size=n)
    bathy = np.ones(n) * 60.2
    sal = np.ones(n) * 0.23

    # gather in a dict
    bathy = {"value":bathy, "lat":lats, "lon":lons}
    sal = {"value":sal, "lat":lats, "lon":lons, "depth":depths}

    # initialize Ocean class
    o = Ocean(load_bathymetry=bathy, load_salinity=sal, **bounds, interp_args={"max_size":150, "method":"linear"})

    # get interpolated values
    b = o.bathymetry(lat=43, lon=-61) #inside data region
    assert np.isclose(b, 60.2, rtol=1e-5)

    b = o.bathymetry(lat=49, lon=-60) #outside data region
    assert np.isclose(b, 60.2, rtol=1e-5)

    s = o.salinity(lat=43, lon=-61, depth=100) #inside data region 
    assert np.isclose(s, 0.23, rtol=1e-5) 

    s = o.salinity(lat=49, lon=-60, depth=10000) #outside data region
    assert np.isclose(s, 0.23, rtol=1e-5)


@pytest.mark.hycom_access
def test_small_full_ocean():
    """ Test that the ocean can be initialized for a very small region 
        using GEBCO bathymetry and HYCOM temperature and salinity
    """

    bounds = dict(
        start=datetime(2019, 1, 9, 0),
        end=datetime(2019, 1, 9, 12), #12 hours -> four 3-hour bins
        south=38.2,
        west=-55.4,
        north=39.3,
        east=-54.3,
        top=0,
        bottom=1
    )

    o = Ocean(
        load_bathymetry='gebco',
        load_temperature='hycom',
        load_salinity='hycom',
        #load_wavedirection='wwiii', #TODO: uncomment once bug in WWIII module has been fixed
        #load_waveheight='wwiii',
        #load_waveperiod='wwiii',
        #load_wind_uv='wwiii',
        **bounds,
        drop=["epoch"]
    )

    # check bathy at one location
    b = o.bathymetry(lat=39, lon=-55)        
    assert b == pytest.approx(5298, abs=10)

    # check salinity at one location, depth and instant in time
    t = datetime(2019, 1, 9, 7) 
    t0 = datetime(2000, 1, 1)
    #epoch = (t - t0).total_seconds() / 3600 #hours since since 2000-01-01 00:00
    s = o.salinity(lat=39, lon=-55, depth=0.5)
    assert s == pytest.approx(36.27, abs=0.05)


"""
    additional tools and utils used across data loading modules
"""

import configparser
import logging
import os
import sqlite3
import warnings
from datetime import datetime, timedelta
from functools import reduce, partial
from os.path import dirname

import numpy as np

ext = lambda filepath, extensions: isinstance(extensions, tuple) and any(
    x.lower() == filepath.lower()[-len(x):] for x in extensions)


def database_cfg():
    """ configure and connect to sqlite database

        time is stored as an integer in the database, where each value
        is epoch hours since 2000-01-01 00:00

        returns:
            conn:
                database connection object
            db:
                connection cursor object
    """
    conn = sqlite3.connect(storage_cfg() + 'geospatial.db')
    db = conn.cursor()

    return conn, db


class storage_cfg(os.PathLike):
    """ return filepath containing storage configuration string

        first checks the config.ini file in kadlu root folder, then
        defaults to kadlu/storage
    """
    """
    __file__  = '/home/matt_s/kadlu/kadlu/geospatial/data_sources/data_util.py'
    """
    cfg = configparser.ConfigParser()  # read .ini into dictionary object

    def __init__(self, setdir=None):
        self.cfg.read(
            os.path.join(dirname(dirname(dirname(dirname(__file__)))),
                         "config.ini"))
        if setdir:
            self.__call__(setdir)

    def default_storage(self, msg):
        """ helper function for storage_cfg() """
        storage_location = os.path.join(os.path.expanduser('~'),
                                        f'kadlu_data{os.path.sep}')
        if not os.path.isdir(storage_location):
            os.mkdir(storage_location)
        warnings.warn(
            f'''{msg} storage location will be set to {storage_location}.\n'''
            '''use kadlu.storage_cfg('/path/to/new/directory') '''
            '''to configure a different storage location''')
        return storage_location

    def __call__(self, setdir=None) -> str:
        if 'storage' not in self.cfg.sections():
            self.cfg.add_section('storage')

        if setdir is not None:
            assert os.path.isdir(setdir), 'directory does not exist'
            self.cfg.set('storage', 'storage_location',
                         os.path.abspath(setdir))
            with open(
                    os.path.join(dirname(dirname(dirname(dirname(__file__)))),
                                 "config.ini"), 'w') as f:
                self.cfg.write(f)
        try:
            storage_location = self.cfg['storage']['storage_location']
        except KeyError:  # missing config.ini file
            return self.default_storage('storage location not configured.')

        if storage_location == '':  # null value in config.ini
            return self.default_storage('null value in config.')

        if not os.path.isdir(storage_location):  # verify the location exists
            return self.default_storage('storage location does not exist.')

        if storage_location[-1] != os.path.sep:
            return storage_location + os.path.sep

        return str(os.path.abspath(storage_location))

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self.__call__()

    def __dir__(self) -> str:
        return self.__call__()

    def __fspath__(self) -> str:
        return self.__call__()

    def __add__(self, other) -> str:
        return ''.join([self.__str__(), other.__str__()])


def verbosity(set_verbosity=None):
    '''
    __file__ = '/home/matt/kadlu/kadlu/geospatial/data_sources/data_util.py'
    '''
    cfg = configparser.ConfigParser()
    cfg.read(
        os.path.join(dirname(dirname(dirname(dirname(__file__)))),
                     "config.ini"))

    if 'other' not in cfg.sections():
        cfg.add_section('other')
        cfg.set('other', 'verbosity', str(False))
        with open(
                os.path.join(dirname(dirname(dirname(dirname(__file__)))),
                             "config.ini"), 'w') as f:
            cfg.write(f)

    if set_verbosity is not None:
        assert set_verbosity in (True,
                                 False), 'verbosity must be True or False'
        cfg.set('other', 'verbosity', str(set_verbosity))
        with open(
                os.path.join(dirname(dirname(dirname(dirname(__file__)))),
                             "config.ini"), 'w') as f:
            cfg.write(f)

    return cfg['other']['verbosity'] == 'True'


def dt_2_epoch(dt_arr, t0=datetime(2000, 1, 1, 0, 0, 0)):
    """ convert datetimes to epoch hours """
    delta = lambda dt: (dt - t0).total_seconds() / 60 / 60
    if isinstance(dt_arr, (list, np.ndarray)):
        return list(map(int, map(delta, dt_arr)))
    elif isinstance(dt_arr, (datetime)):
        return int(delta(dt_arr))
    else:
        raise ValueError('input must be datetime or array of datetimes')


def epoch_2_dt(ep_arr, t0=datetime(2000, 1, 1, 0, 0, 0), unit='hours'):
    """ convert epoch hours to datetimes """
    delta = lambda ep, unit: t0 + timedelta(**{f'{unit}': ep})
    if isinstance(ep_arr, (list, np.ndarray)):
        return list(map(partial(delta, unit=unit), ep_arr))
    elif isinstance(ep_arr, (float, int)):
        return delta(ep_arr, unit=unit)
    else:
        raise ValueError('input must be integer or array of integers')


def index_arr(val, sorted_arr):
    """ converts value in coordinate array to grid index """
    if val > sorted_arr[-1]:
        return len(sorted_arr) - 1
    return np.nonzero(sorted_arr >= val)[0][0]


def flatten(cols, frames):
    """ dimensional reduction by taking average of time frames """
    # assert that frames are of equal size
    assert len(cols) == 5
    assert reduce(lambda a, b: (a == b) * a, frames[1:] - frames[:-1])

    # aggregate time frame splits and reduce them to average
    frames_ix = np.append([0], frames)
    split_num = range(len(frames_ix) - 1)
    val_split = np.array(
        [cols[0][frames_ix[f]:frames_ix[f + 1]] for f in split_num])
    value_avg = (reduce(np.add, val_split) / len(val_split))
    _, y, x, _, z = cols[:, frames_ix[0]:frames_ix[1]]
    return value_avg, y, x, z


def reshape_2D(cols):
    return dict(values=cols[0], lats=cols[1], lons=cols[2])


def reshape_3D(cols):
    return dict(values=cols[0], lats=cols[1], lons=cols[2], depths=cols[3])


def reshape_4D(cols):
    """ prepare loaded data for interpolation

        TODO: review, validate and cleanup this function
                especially the replacement of nan values!

        args:
            cols: flattened numpy array of shape (5, n)
                cols[0]: values
                cols[1]: latitude
                cols[2]: longitude
                cols[3]: depth
                cols[4]: time (hours since )

        return: gridded
            dict(values=gridspace, lats=ygrid, lons=xgrid, depths=zgrid, times=tgrid)

    """
    if isinstance(cols[0], (float, int)):
        return dict(values=cols[0])

    vals, y, x, t, z = cols

    rows = np.array((vals, y, x, z, t)).T

    # reshape row data to 4D array
    xgrid, ygrid, zgrid, tgrid = np.unique(x), np.unique(y), np.unique(z), np.unique(t)
    gridspace = np.full((len(ygrid), len(xgrid), len(zgrid), len(tgrid)),
                        fill_value=None,
                        dtype=float)

    # this could potentially be optimized to avoid an index lookup cost
    for row in rows:
        x_ix = index_arr(row[2], xgrid)
        y_ix = index_arr(row[1], ygrid)
        z_ix = index_arr(row[3], zgrid)
        t_ix = index_arr(row[4], tgrid)
        gridspace[y_ix, x_ix, z_ix, t_ix] = row[0]

    # remove nulls for interpolation:
    # fill missing depth values with last value in each column
    for xi in range(0, gridspace.shape[0]):
        for yi in range(0, gridspace.shape[1]):
            col = gridspace[xi, yi]
            if np.sum(np.isnan(col)) > 0 and np.sum(np.isnan(col)) < len(col):
                col[np.isnan(col)] = col[~np.isnan(col)][-1]
                gridspace[xi, yi] = col

    # null depth columns are filled with the average value at each depth plane
    for zi in range(0, gridspace.shape[2]):
        gridspace[:, :, zi, :][np.isnan(gridspace[:, :, zi, :])] = np.average(
            gridspace[:, :, zi, :][~np.isnan(gridspace[:, :, zi, :])])

    return dict(values=gridspace, lats=ygrid, lons=xgrid, depths=zgrid, times=tgrid)


def reshape_3D_gridded(cols):
    """ prepare loaded data for interpolation

        TODO: review and cleanup this function

        args:
            cols: flattened numpy array of shape (4, n)
                cols[0]: values
                cols[1]: latitude
                cols[2]: longitude
                cols[3]: depth

        return: gridded
            dict(values=gridspace, lats=ygrid, lons=xgrid, depths=zgrid)

    """
    if isinstance(cols[0], (float, int)):
        return dict(values=cols[0])

    # if data is 4D, take average of values at time frame intervals
    frames = np.append(
        np.nonzero(cols[3][1:] > cols[3][:-1])[0] + 1, len(cols[3])
    )

    if len(np.unique(frames)) > 1:
        vals, y, x, z = flatten(cols, frames)
    else:
        vals, y, x, _, z = cols

    rows = np.array((vals, y, x, z)).T

    # reshape row data to 3D array
    xgrid, ygrid, zgrid = np.unique(x), np.unique(y), np.unique(z)
    gridspace = np.full((len(ygrid), len(xgrid), len(zgrid)),
                        fill_value=None,
                        dtype=float)

    # this could potentially be optimized to avoid an index lookup cost
    for row in rows:
        x_ix = index_arr(row[2], xgrid)
        y_ix = index_arr(row[1], ygrid)
        z_ix = index_arr(row[3], zgrid)
        gridspace[y_ix, x_ix, z_ix] = row[0]

    # remove nulls for interpolation:
    # fill missing depth values with last value in each column
    for xi in range(0, gridspace.shape[0]):
        for yi in range(0, gridspace.shape[1]):
            col = gridspace[xi, yi]
            if sum(np.isnan(col)) > 0 and sum(np.isnan(col)) < len(col):
                col[np.isnan(col)] = col[~np.isnan(col)][-1]
                gridspace[xi, yi] = col

    # null depth columns are filled with the average value at each depth plane
    for zi in range(0, gridspace.shape[2]):
        gridspace[:, :, zi][np.isnan(gridspace[:, :, zi])] = np.average(
            gridspace[:, :, zi][~np.isnan(gridspace[:, :, zi])])

    return dict(values=gridspace, lats=ygrid, lons=xgrid, depths=zgrid)


def str_def(self, info, args):
    """ builds string definition for data source class objects """
    fcns = [
        fcn for fcn in dir(self) if callable(getattr(self, fcn))
        and 'load' in fcn and not fcn.startswith("__")
    ]
    return (
        f'{info}\n\nfunction input arguments:\n\t{args}\n\nclass functions:'
        '\n\t' + '\n\t'.join(fcns) + '\n')


def ll_2_regionstr(south, north, west, east, regions, default=[]):
    """ convert input bounds to region strings with seperating axis theorem """
    if west > east:  # recursive function call if query intersects antimeridian
        return np.union1d(
            ll_2_regionstr(south, north, west, +180, regions, default),
            ll_2_regionstr(south, north, -180, east, regions, default))

    query = Boundary(south, north, west, east)
    matching = [str(reg) for reg in regions if query.intersects(reg)]

    if len(matching) == 0:
        warnings.warn("No regions matched for query. "
                      f"Defaulting to {default} ({len(default)} regions)")
        return default

    return np.unique(matching)


def fmt_coords(kwargs):
    """Formats spatial coordinates as a human readable character string
    
    Args:
        kwargs: dict
            Must have keys `south`, `north`, `east`, `west` and optionally also `top` and `bottom`

    Returns:
        : str
            Nicely formatted string
    """
    kwargs = kwargs.copy()

    # if longitudes are in range [0;360] convert to [-180;180]
    if kwargs['east'] > 180:
        kwargs['east'] -= 360.

    if kwargs['west'] > 0:
        kwargs['west'] -= 360.

    return (
        f"{abs(kwargs['south']):.2f}°{'S' if kwargs['south'] <= 0 else 'N'},"
        f"{abs(kwargs['north']):.2f}°{'S' if kwargs['north'] <= 0 else 'N'};"
        f"{abs(kwargs['west']):.2f}°{'W' if kwargs['west'] <= 0 else 'E'}"
        f"{abs(kwargs['east']):.2f}°{'W' if kwargs['east'] <= 0 else 'E'}"
        f"{','+str(kwargs['bottom'])+'m' if 'bottom' in kwargs.keys() else ''}"
        f"{','+str(kwargs['top'])+'m' if 'top' in kwargs.keys() else ''}")


def fmt_time(kwargs):
    """Formats time window as a human readable character string
    
    Args:
        kwargs: dict
            Must have keys `start` and `end`

    Returns:
        : str
            Nicely formatted string
    """
    return f"{kwargs['start']} to {kwargs['end']}"


def logmsg(source, var, ntup=(), **kwargs):
    """ Log message informing that data was inserted into the kadlu database    

        The message includes
            
            * the name of the data source
            * the name of the variable 
            * the number of data points added to the database
            * the geographical region
            * the time period 

        Args:
            source: str
                Data source name
            var: str
                Variable name
            ntup: tuple
                ... If None, the message will state that no data was found.
        
        Keyword args:
            south,north,east,west: float
                Geographic region
            top,bottom: float
                Depth range
            start,end: datetime.datetime
                Time window            
    """
    info_msg = f"{source.upper()} {var.lower()}: "

    if ntup is None:
        info_msg += "no data found, returning empty arrays"
    else:
        info_msg += f"logged {reduce(np.subtract, ntup[::-1])} values"

    if all(x in kwargs for x in ["south","north","east","west"]): 
        info_msg += f" in region {fmt_coords(kwargs)}"

    if all(x in kwargs for x in ["start","end"]):
        info_msg += f" for time period {fmt_time(kwargs)}"
                 
    logging.info(info_msg)


def logmsg_nodata(source, var, **kwargs):
    """ Log message informing that data was inserted in the database
    """
    return logmsg(source, var, ntup=None, **kwargs)


class Boundary():
    """ compute intersecting boundaries with separating axis theorem """

    def __init__(self, south, north, west, east, fetchvar='', **_):
        self.south = south
        self.north = north
        self.west = west
        self.east = east
        self.fetchvar = fetchvar

    def __str__(self):
        return self.fetchvar

    def intersects(self, other):
        return not (self.east < other.west or self.west > other.east
                    or self.north < other.south or self.south > other.north)



import os
from pathlib import Path
from setuptools import setup, find_packages

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='kadlu',
    version=os.environ.get('KADLUVERSION', '2.4.1'),
    description="MERIDIAN Python package for ocean ambient noise modelling",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.meridian.cs.dal.ca/public_projects/kadlu',
    author='Oliver Kirsebom, Matthew Smith',
    author_email='oliver.kirsebom@dal.ca, matthew.smith@dal.ca',
    license='GNU General Public License v3.0',
    packages=find_packages(exclude=('tests', )),
    install_requires=[
        'Pillow',
        'cdsapi',
        'eccodes',
        'gsw',
        'imageio',
        'matplotlib',
        'mpl_scatter_density',
        'netcdf4',
        'numpy',
        'pygrib',  # DEPENDS ON eccodes binaries
        'pyproj',
        #'pyqt5',
        'requests',
        'scipy',
        'tqdm',
        'xarray',
    ],
    #setup_requires=[ 'pytest-runner', ],
    #tests_require=['pytest', 'pytest-parallel'],
    tests_require=['pytest'],
    include_package_data=True,
    exclude_package_data={'': ['tests']},
    python_requires='>=3.6',
)



import os
import logging
import zipfile
import requests
from datetime import datetime, timedelta
import time

import numpy as np
from tqdm import tqdm

import kadlu
from kadlu import index
from kadlu.geospatial.data_sources.data_util import (
    database_cfg,
    logmsg,
    logmsg_nodata,
    storage_cfg,
)

zipfilename = 'gebco_2021_geotiff.zip'
url = 'https://www.bodc.ac.uk/data/open_download/gebco/gebco_2021/geotiff/'


def initdb():
    conn, db = database_cfg()
    db.execute('CREATE TABLE IF NOT EXISTS gebco'
               '(   val     REAL    NOT NULL,  '
               '    lat     REAL    NOT NULL,  '
               '    lon     REAL    NOT NULL  )')
    db.execute(
        'CREATE UNIQUE INDEX IF NOT EXISTS idx_gebco on gebco(val, lat, lon)')
    conn.close()


class Gebco():

    def fetch_bathymetry_grid(self):
        """ download geotiff zip archive and extract it """

        zipf = os.path.join(storage_cfg(), "gebco_2021_geotiff.zip")

        # download the file if necessary
        if not os.path.isfile(zipf):
            print('downloading gebco bathymetry...')
            with requests.get(url, stream=True, timeout=300) as payload:
                assert payload.status_code == 200, 'error fetching file'
                with open(zipf, 'wb') as f:
                    with tqdm(total=4011407575,
                              desc=zipf,
                              unit='B',
                              unit_scale=True) as t:
                        for chunk in payload.iter_content(chunk_size=8192):
                            _ = t.update(f.write(chunk))

            # unzip the downloaded file
            exists = set(sorted(os.listdir(storage_cfg())))
            with zipfile.ZipFile(zipf, 'r') as zip_ref:
                contents = set(zip_ref.namelist())
                members = list(contents - exists)
                print('extracting bathymetry data...')
                zip_ref.extractall(path=storage_cfg(), members=members)

        time.sleep(2)

        return

    def fetch_callback(self,
                       south,
                       north,
                       west,
                       east,
                       top=None,
                       bottom=None,
                       start=None,
                       end=None):
        """ build grid indexes from geotiffs and insert into the database """

        if not os.path.isfile(os.path.join(storage_cfg(), zipfilename)):
            self.fetch_bathymetry_grid()

        filebounds = lambda fpath: {
            f[0]: float(f[1:])
            for f in fpath.split('gebco_2021_', 1)[1].rsplit('.tif', 1)[0].
            split('_')
        }

        rasterfiles = {
            f: filebounds(f)
            for f in {
                k: None
                for k in sorted([
                    f for f in os.listdir(storage_cfg())
                    if f[-4:] == '.tif' and 'gebco_2021' in f
                ])
            }
        }

        initdb()
        for fname, fbounds in rasterfiles.items():
            in_lat_rng = (fbounds['s'] <= south <= fbounds['n']
                          or fbounds['s'] <= north <= fbounds['n'])
            in_lon_rng = (fbounds['w'] <= west <= fbounds['e']
                          or fbounds['w'] <= east <= fbounds['e'])
            if (not in_lat_rng or not in_lon_rng):
                continue
            assert os.path.isfile(os.path.join(storage_cfg(), fname))
            grid = kadlu.load_file(os.path.join(storage_cfg(), fname),
                                   south=south,
                                   north=north,
                                   west=west,
                                   east=east).T

            conn, db = database_cfg()
            n1 = db.execute("SELECT COUNT(*) FROM gebco ").fetchall()[0][0]
            db.executemany(
                "INSERT OR IGNORE INTO gebco (val, lat, lon)  VALUES (?,?,?)",
                grid)
            n2 = db.execute("SELECT COUNT(*) FROM gebco ").fetchall()[0][0]

            logmsg('gebco',
                   'bathymetry', (n1, n2),
                   south=south,
                   west=west,
                   north=north,
                   east=east)
            conn.commit()
            conn.close()
        return

    def load_bathymetry(self, south, north, west, east, **_):
        """ load gebco bathymetry data """

        with index(dx=2,
                   dy=2,
                   dz=99999,
                   dt=timedelta(hours=24),
                   storagedir=storage_cfg(),
                   bins=True,
                   south=south,
                   north=north,
                   west=west,
                   east=east,
                   top=0,
                   bottom=99999,
                   start=datetime(2000, 1, 1),
                   end=datetime(2000, 1, 2)) as fetchmap:
            _ = fetchmap(callback=self.fetch_callback)

        conn, db = kadlu.database_cfg()
        qry = ' AND '.join([
            'SELECT val, lat, lon FROM gebco WHERE lat >= ?', 'lat <= ?',
            'lon >= ?', 'lon <= ?'
        ])
        logging.debug(f'query: {qry}')
        db.execute(qry, (south, north, west, east))
        res = np.array(db.fetchall()).T
        conn.close()
        if len(res) == 0:
            logmsg_nodata('gebco',
                          'bathymetry',
                          south=south,
                          north=north,
                          west=west,
                          east=east)
            return np.array([[], [], []])
        val, lat, lon = res
        return np.array([val * -1, lat, lon])



import kadlu
from datetime import datetime

# gulf st lawrence
start = datetime(2014, 2, 3, 0, 0, 0, 0)
end = datetime(2014, 2, 3, 3, 0, 0, 0)
y, x, offset = 47.133, -63.51, 1
south, west = y - offset, x - offset
north, east = y + offset, x + offset


def test_wwiii_load_windwaveheight():
    result = kadlu.load(source='wwiii',
                        var='waveheight',
                        south=south,
                        west=west,
                        north=north,
                        east=east,
                        start=start,
                        end=end)


def test_wwiii_load_wavedir():
    result = kadlu.load(source='wwiii',
                        var='wavedir',
                        south=south,
                        west=west,
                        north=north,
                        east=east,
                        start=start,
                        end=end)


def test_load_wind():
    ns_offset = 1
    ew_offset = 1

    uvval, lat, lon, epoch = kadlu.load(source='wwiii',
                                        var='wind_uv',
                                        start=datetime(2016, 3, 9),
                                        end=datetime(2016, 3, 11),
                                        south=44.5541333 - ns_offset,
                                        west=-64.17682 - ew_offset,
                                        north=44.5541333 + ns_offset,
                                        east=-64.17682 + ew_offset,
                                        top=0,
                                        bottom=0)

    assert (len(uvval) == len(lat) == len(lon))
    assert len(uvval) > 0



""" Sound speed module within the kadlu package
"""
import gsw
import numpy as np
import logging
from datetime import datetime
from kadlu.utils import interp_grid_1d
from kadlu.geospatial.interpolation import get_interpolator


def sound_speed_teos10(lats, lons, z, t, SP):
    """ Compute sound speed from temperature, salinity, and depth
        for a given latitude and longitude using the Thermodynamic
        Equation of Seawater 2010.

        https://teos-10.github.io/GSW-Python/

        Args:
            lats: numpy array
                Latitudes (-90 to 90 degrees)
            lons: numpy array
                Longitudes (-180 to 180 degrees)
            z: numpy array
                Depths (meters)
            t: numpy array
                In-situ temperature (Celsius)
            SP: numpy array
                Practical Salinity (psu)

        Returns:
            c: numpy array
                Sound speed (m/s)
    """
    p = gsw.p_from_z(
        z=-z, lat=lats)  # sea pressure (gsw uses negative z below sea surface)
    SA = gsw.SA_from_SP(SP, p, lons, lats)  # absolute salinity
    CT = gsw.CT_from_t(SA, t, p)  # conservative temperature
    c = gsw.density.sound_speed(SA=SA, CT=CT, p=p)
    return c


class SoundSpeed():
    """ Class for handling computation and interpolation of sound speed.

        The sound speed can be specified via the argument ssp (sound speed
        profile) or computed from the ocean variables (temperature, salinity).

        ssp can be either a single value, in which case the sound speed is
        the same everywhere, or a tuple (c,z) where c is an array of sound
        speed values and z is an array of depths.

        The interp and interp_xyz method may be used to obtain the interpolated
        sound speed at any set of coordinates.

        TODO: provide proper implementation of time/epoch handling.
        TODO: SoundSpeed should inherit from BaseGeospatialInterpolator

        Args:
            ocean: instance of :class:`kadlu.geospatial.ocean.Ocean`
                Ocean variables
            ssp: float or tuple
                Sound speed profile. May be specified either as a float,
                in which case the sound speed is the same everywhere, or as
                a tuple (c,z) where c is an array of sound speeds and z is
                an array of depth values.
            num_depths: int
                Number of depth values for the interpolation grid. The default value is 50.
            rel_err: float
                Maximum deviation of the interpolation, expressed as a ratio of the
                range of sound-speed values. The default value is 0.001.
    """

    def __init__(self, ocean=None, ssp=None, num_depths=50, rel_err=1E-3, time=None):

        assert ocean is not None or ssp is not None, "ocean or ssp must be specified"

        self.logger = logging.getLogger("kadlu")
        self.name = self.__class__.__name__

        if ssp is not None:
            self.logger.debug(f"[{self.name}] Initializing {self.name} instance with SSP = {ssp}")

            if isinstance(ssp, tuple):
                self._interp = get_interpolator(value=ssp[0], depth=ssp[1])
            else:
                self._interp = get_interpolator(value=ssp)

        else:
            self.logger.debug(f"[{self.name}] Initializing {self.name} instance with Ocean data")

            lat_res, lon_res = self._lat_lon_res(
                ocean, default_res=1.0
            )  #default resolution is 1 degree, approx 100 km

            # geographic boundaries
            S, N, W, E = ocean.boundaries['south'], ocean.boundaries[
                'north'], ocean.boundaries['west'], ocean.boundaries['east']

            # patch: ok 2024-07-17
            # only appropriate for HYCOM data!
            if time is None:
                time = ocean.boundaries['start']

            t0 = datetime(2000, 1, 1)
            epoch = (time - t0).total_seconds() / 3600 #hours since since 2000-01-01 00:00

            # lat and lon coordinates
            num_lats = max(3, int(np.ceil((N - S) / lat_res)) + 1)
            lats = np.linspace(S, N, num=num_lats)
            num_lons = max(3, int(np.ceil((E - W) / lon_res)) + 1)
            lons = np.linspace(W, E, num=num_lons)

            # compute depth coordinates
            depths = self._depth_coordinates(
                ocean,
                lats,
                lons,
                num_depths=num_depths,
                rel_err=rel_err,
                epoch=epoch,
            )

            debug_msg = f"[{self.name}] Interpolation sound speed on grid with:"
            debug_msg += f"\n[{self.name}] - resolution: lat,lon={lat_res:.4f},{lon_res:.4f}"
            debug_msg += f"\n[{self.name}] - size: lats,lons,depths={num_lats},{num_lons},{num_depths}"
            debug_msg += f"\n[{self.name}] - extent: lat,lon,depth=({np.min(lats):.4f},{np.max(lats):.4f})"
            debug_msg += f",({np.min(lons):.4f},{np.max(lons):.4f})"
            debug_msg += f",({np.min(depths):.1f},{np.max(depths):.1f})"
            self.logger.debug(debug_msg)

            # interpolate temperature and salinity on lat,lon,epoch,depth grid
            t = ocean.temperature(lat=lats, lon=lons, depth=depths, epoch=epoch, grid=True)
            s = ocean.salinity(lat=lats, lon=lons, depth=depths, epoch=epoch, grid=True)

            # drop time axis with size 1
            t = t[:,:,0,:]
            s = s[:,:,0,:]

            # compute sound speed
            grid_shape = t.shape
            la, lo, de = np.meshgrid(lats, lons, depths)
            la = la.flatten()
            lo = lo.flatten()
            de = de.flatten()
            t = t.flatten()
            s = s.flatten()
            c = sound_speed_teos10(lats=la, lons=lo, z=de, t=t, SP=s)
            c = np.reshape(c, newshape=grid_shape)

            # create interpolator
            self._interp = get_interpolator(value=c, lat=lats, lon=lons, depth=depths, origin=ocean.origin, method='linear')

    def _lat_lon_res(self, ocean, default_res, min_res=1E-5):
        """ Determine lat,lon resolutions for interpolation grid

            Args:
                ocean: instance of :class:`kadlu.geospatial.ocean.Ocean`
                    Ocean variables
                default_res: float
                    Default resolution in degrees
                min_res: float
                    Highest resolution in degrees. Default is 1E-5, corresponding to approx 1 meter.

            Returns:
                lat_res, lon_res: float,float
                    Resolutions in degrees.
        """
        temp_nodes = ocean.interpolators['temperature'].coordinates
        temp_lat, temp_lon = temp_nodes.get("lat",None), temp_nodes.get("lon",None)

        sal_nodes = ocean.interpolators['salinity'].coordinates
        sal_lat, sal_lon = sal_nodes.get("lat",None), sal_nodes.get("lon",None)

        if temp_lat is None or len(temp_lat) <= 1:
            temp_lat_res = default_res
        else:
            temp_lat_res = max(min_res, np.abs(temp_lat[1] - temp_lat[0]))

        if temp_lon is None or len(temp_lon) <= 1:
            temp_lon_res = default_res 
        else:
            temp_lon_res = max(min_res, np.abs(temp_lon[1] - temp_lon[0]))

        if sal_lat is None or len(sal_lat) <= 1:
            sal_lat_res = default_res
        else:
            sal_lat_res = max(min_res, np.abs(sal_lat[1] - sal_lat[0]))

        if sal_lon is None or len(sal_lon) <= 1:
            sal_lon_res = default_res
        else:
            sal_lon_res = max(min_res, np.abs(sal_lon[1] - sal_lon[0]))

        lat_res = min(temp_lat_res, sal_lat_res)
        lon_res = min(temp_lon_res, sal_lon_res)

        return lat_res, lon_res

    def _depth_coordinates(self, ocean, lats, lons, num_depths, rel_err, epoch):
        """ Compute depth coordinates for lat,lon,depth interpolation grid.

            Args:
                ocean: instance of :class:`kadlu.geospatial.ocean`
                    Ocean variables
                lats,lons: numpy.array, numpy.array
                    Latitude and longitude coordinates
                num_depths: int
                    Number of depth values for the interpolation grid. The default value is 50.
                rel_err: float
                    Maximum deviation of the interpolation, expressed as a ratio of the
                    range of sound-speed values. The default value is 0.001.

            Returns:
                depths: numpy.array
                    Depth coordinates
        """
        seafloor_depth = ocean.bathymetry(lat=lats, lon=lons, grid=True)

        # find deepest point
        deepest_point = np.unravel_index(np.argmax(seafloor_depth),
                                         seafloor_depth.shape)

        # depth and lat,lon coordinates at deepest point
        max_depth = seafloor_depth[deepest_point]
        lat = lats[deepest_point[0]]
        lon = lons[deepest_point[1]]

        self.logger.debug(f"[{self.name}] Deepest point: lat,lon,depth={lat:.4f},{lon:.4f},{max_depth:.1f}m")

        # compute temperature, salinity and sound speed for every 1 meter
        z = np.arange(0, int(np.ceil(max_depth)) + 1)
        t = ocean.temperature(lat=lat, lon=lon, depth=z, epoch=epoch, grid=True)
        s = ocean.salinity(lat=lat, lon=lon, depth=z, epoch=epoch, grid=True)
        c = sound_speed_teos10(lats=lat, lons=lon, z=z, t=t, SP=s)

        # determine grid
        indices, _ = interp_grid_1d(
            y=np.squeeze(c),
            x=z,
            num_pts=num_depths,
            rel_err=rel_err
        )

        depths = z[indices]

        return depths


    def __call__(self, **kwargs):
        return self._interp(**kwargs)
